/**
  ******************************************************************************
  * @file           : mpu9250.c
  * @brief          : MPU9250 driver
  * @author         : MicroTechnics (microtechnics.ru)
  *                   based on the work of Kris Winer
  *                   https://github.com/kriswiner/MPU9250
  ******************************************************************************
  */



/* Includes ------------------------------------------------------------------*/

#include "MPU9250.h"



/* Declarations and definitions ----------------------------------------------*/

#define Kp 2.0f * 5.0f
#define Ki 0.0f

enum Ascale {
  AFS_2G = 0,
  AFS_4G,
  AFS_8G,
  AFS_16G
};

enum Gscale {
  GFS_250DPS = 0,
  GFS_500DPS,
  GFS_1000DPS,
  GFS_2000DPS
};

enum Mscale {
  MFS_14BITS = 0,
  MFS_16BITS
};

MPU9250_CalibrationStatus calibStatus = MPU9250_CALIBRATION_NOT_ACTIVE;

uint8_t Ascale = AFS_2G;
uint8_t Gscale = GFS_250DPS;
uint8_t Mscale = MFS_16BITS;

uint8_t Mmode = 0x06;
float aRes, gRes, mRes;

int16_t accelCount[AXIS_NUM];
int16_t gyroCount[AXIS_NUM];
int16_t magCount[AXIS_NUM];

float magCalibration[AXIS_NUM] = {0, 0, 0};
float magBias[AXIS_NUM] = {0, 0, 0};
float magScale[AXIS_NUM] = {1, 1, 1};
float gyroBias[AXIS_NUM] = {0, 0, 0};
float gyroScale[AXIS_NUM] = {1, 1, 1};
float accelBias[AXIS_NUM] = {0, 0, 0};
float accelScale[AXIS_NUM] = {1, 1, 1};

float ax, ay, az, gx, gy, gz, mx, my, mz;

int16_t tempCount;
float temperature;
float SelfTest[6];
int delt_t = 0;
int count = 0;

float PI = 3.14159265358979323846f;
float beta = 0.86602540378443864676f * 3.14159265358979323846f * (60.0f / 180.0f);

float MPU9250_CompassAngles[ANGLES_NUM];
float MPU9250_Acc[AXIS_NUM];
float MPU9250_Gyro[AXIS_NUM];

float deltat = 0.0f;
int lastUpdate = 0, firstUpdate = 0, Now = 0;

float q[4] = {1.0f, 0.0f, 0.0f, 0.0f};
float eInt[3] = {0.0f, 0.0f, 0.0f};
float mpuSelfTesf[6];
float sum = 0;
uint32_t sumCount = 0;
float deltaWatch = 0;
uint8_t calibration = 0;

float accPitch = 0;
float accRoll = 0;

float accMax[AXIS_NUM] = {0, 0, 0};
float accMin[AXIS_NUM] = {0, 0, 0};

float xHorFilt = 0;
float yHorFilt = 0;

float mxHorFilt = 0;
float myHorFilt = 0;

float coef = 0.8;
float heading = 0;

float calibSum[AXIS_NUM] = {0, 0, 0};

extern int uSec;
extern I2C_HandleTypeDef hi2c1;

static int32_t calibMagBias[AXIS_NUM] = {0, 0, 0};
static int32_t calibMagScale[AXIS_NUM] = {0, 0, 0};

static int16_t calibMagMax[AXIS_NUM] = {-32768, -32768, -32768};
static int16_t calibMagMin[AXIS_NUM] = {32767, 32767, 32767};
static int16_t calibMagTemp[AXIS_NUM] = {0, 0, 0};

static uint16_t calibStep = 0;
static uint16_t calibSamplesNum = 0;

void getMres();
void getAres();
void getGres();
void readMagData(int16_t * destination);
void readAccelData(int16_t * destination);
void readGyroData(int16_t * destination);
char readByte(uint8_t address, uint8_t subAddress);
void resetMPU9250();
int16_t readTempData();
void initAK8963(float * destination);
void initMPU9250();



/* Functions -----------------------------------------------------------------*/

/**
  * @brief  Writes data to sensor's registers
  * @param  devAdrress device address
  * @param  regAddress register address
  * @param  regData registers' data
  * @param  cnt number of registers to write
  * @retval Status
  */
MPU9250_Status MPU9250_I2C_Write(uint8_t devAdrress, uint8_t regAddress, uint8_t *regData, uint8_t cnt)
{
  HAL_StatusTypeDef i2cStatus = HAL_OK;
  
  if ((cnt + 1) > MPU9250_I2C_MAX_DATA_LEN)
  {
    return MPU9250_I2C_OVERFLOW;
  }
  
  uint8_t tempData[MPU9250_I2C_MAX_DATA_LEN];
  
  tempData[0] = regAddress;

  for (uint8_t i = 0; i < cnt; i++)
  {
    tempData[i + 1] = *regData;
    regData++;
  }
  
  i2cStatus = HAL_I2C_Master_Transmit(&hi2c1, (devAdrress << 1), tempData, (cnt + 1),  MPU9250_I2C_TIMEOUT);

  if (i2cStatus != HAL_OK)
  {
    return MPU9250_ERROR;
  }

  return MPU9250_OK;
}



/**
  * @brief  Reads data from sensor's registers
  * @param  devAdrress device address
  * @param  regAddress register address
  * @param  regData pointer for data storing
  * @param  cnt number of registers to read
  * @retval Status
  */
MPU9250_Status MPU9250_I2C_Read(uint8_t devAdrress, uint8_t regAddress, uint8_t *regData, uint8_t cnt)
{
  HAL_StatusTypeDef i2cStatus = HAL_OK;
  
  i2cStatus = HAL_I2C_Master_Transmit(&hi2c1, (devAdrress << 1), &regAddress, 1,  MPU9250_I2C_TIMEOUT);

  if (i2cStatus != HAL_OK)
  {
    return MPU9250_ERROR;
  }
  
  i2cStatus = HAL_I2C_Master_Receive(&hi2c1, (devAdrress << 1), regData, cnt,  MPU9250_I2C_TIMEOUT);

  if (i2cStatus != HAL_OK)
  {
    return MPU9250_ERROR;
  } 

  return MPU9250_OK;
}



void MPU9250_CalibrationStart(MPU9250_Sensor sensor)
{
  switch (sensor)
  {
    case MPU9250_MAG_SENSOR:
      calibStatus = MPU9250_MAG_CALIBRATION;
      calibSamplesNum = MPU9250_MAG_CALIB_STEPS_NUM;
      break;
      
    case MPU9250_ACC_SENSOR:
      calibStatus = MPU9250_ACC_CALIBRATION;
      calibSamplesNum = MPU9250_ACC_CALIB_STEPS_NUM;
      break;
      
    case MPU9250_GYRO_SENSOR:
      calibStatus = MPU9250_GYRO_CALIBRATION;
      calibSamplesNum = MPU9250_GYRO_CALIB_STEPS_NUM;
      break;
      
    default:
      return;
      break;
  }   
    
  calibStep = 0;
  
  for (uint8_t i = 0; i < AXIS_NUM; i++)
  {
    calibMagMin[i] = 32767;
    calibMagMax[i] = -32768;
    calibMagBias[i] = 0;
    calibMagScale[i] = 0;
    calibMagTemp[i] = 0;
    calibSum[i] = 0;
  }
  
  getMres();
  getAres();
  getGres();
}



void MPU9250_CalibrationFinish()
{
  if (calibStatus == MPU9250_MAG_CALIBRATION)
  {
    calibMagBias[0]  = (calibMagMax[0] + calibMagMin[0]) / 2;
    calibMagBias[1]  = (calibMagMax[1] + calibMagMin[1]) / 2;
    calibMagBias[2]  = (calibMagMax[2] + calibMagMin[2]) / 2;

    magBias[0] = (float)calibMagBias[0] * mRes * magCalibration[0];
    magBias[1] = (float)calibMagBias[1] * mRes * magCalibration[1];
    magBias[2] = (float)calibMagBias[2] * mRes * magCalibration[2];

    calibMagScale[0]  = (calibMagMax[0] - calibMagMin[0]) / 2;
    calibMagScale[1]  = (calibMagMax[1] - calibMagMin[1]) / 2;
    calibMagScale[2]  = (calibMagMax[2] - calibMagMin[2]) / 2;

    float avg_rad = calibMagScale[0] + calibMagScale[1] + calibMagScale[2];
    avg_rad /= 3.0;

    magScale[0] = avg_rad / ((float)calibMagScale[0]);
    magScale[1] = avg_rad / ((float)calibMagScale[1]);
    magScale[2] = avg_rad / ((float)calibMagScale[2]);
    
    FLASH_WriteMagCalibData(magBias, magScale);
  }
  
  if (calibStatus == MPU9250_ACC_CALIBRATION)
  {
    accelBias[0] = calibSum[0] / calibSamplesNum;
    accelBias[1] = calibSum[1] / calibSamplesNum;
    accelBias[2] = calibSum[2] / calibSamplesNum - 1;
    
    accelScale[0] = 1;
    accelScale[1] = 1;
    accelScale[2] = 1;
    
    FLASH_WriteAccCalibData(accelBias, accelScale);
  }
  
  if (calibStatus == MPU9250_GYRO_CALIBRATION)
  {
    gyroBias[0] = calibSum[0] / calibSamplesNum;
    gyroBias[1] = calibSum[1] / calibSamplesNum;
    gyroBias[2] = calibSum[2] / calibSamplesNum;
    
    gyroScale[0] = 1;
    gyroScale[1] = 1;
    gyroScale[2] = 1;
    
    FLASH_WriteGyroCalibData(gyroBias, gyroScale);
  }
  
  calibStep = 0;
  calibSamplesNum = 0;
  calibStatus = MPU9250_CALIBRATION_NOT_ACTIVE;
}



uint8_t MPU9250_CalibrationStep()
{
  uint8_t upd = 0;
    
  if (calibStatus == MPU9250_MAG_CALIBRATION)
  {
    if(readByte(MPU9250_MAGN_I2C_ADDRESS, ST1) & 0x01) 
    {
      readMagData(calibMagTemp); 

      for (int jj = 0; jj < 3; jj++)
      {
        if (calibMagTemp[jj] > calibMagMax[jj])
        {
          calibMagMax[jj] = calibMagTemp[jj];
        }
        
        if (calibMagTemp[jj] < calibMagMin[jj])
        {
          calibMagMin[jj] = calibMagTemp[jj];
        }
      }
      
      upd = 1;
    }
  }
  
  if (calibStatus == MPU9250_ACC_CALIBRATION)
  {
    if (readByte(MPU9250_I2C_ADDRESS, INT_STATUS) & 0x01) 
    {
      readAccelData(accelCount);

      for (uint8_t i = 0; i < AXIS_NUM; i++)
      {
        calibSum[i] += (float)accelCount[i] * aRes;
      }
      
      upd = 1;
    }
  }
  
  if (calibStatus == MPU9250_GYRO_CALIBRATION)
  {
    if (readByte(MPU9250_I2C_ADDRESS, INT_STATUS) & 0x01) 
    {
      readGyroData(gyroCount);

      for (uint8_t i = 0; i < AXIS_NUM; i++)
      {
        calibSum[i] += (float)gyroCount[i] * gRes;
      }
      
      upd = 1;
    }
  }
  
  return upd;
}



void MPU9250_Init()
{
  uint8_t whoami = readByte(MPU9250_I2C_ADDRESS, WHO_AM_I);
  
  if (whoami == 0x71)
  {
    resetMPU9250();    
    initMPU9250();    
    initAK8963(magCalibration);
    
    getAres();
    getGres();
    getMres();
    
    FLASH_ReadCalibData(MPU9250_MAG_SENSOR, magBias, magScale);
    FLASH_ReadCalibData(MPU9250_ACC_SENSOR, accelBias, accelScale);
    FLASH_ReadCalibData(MPU9250_GYRO_SENSOR, gyroBias, gyroScale);
  }
}



void MPU9250_Process()
{
  if (calibStatus == MPU9250_CALIBRATION_NOT_ACTIVE)
  {
    if(readByte(MPU9250_I2C_ADDRESS, INT_STATUS) & 0x01) 
    {
      readAccelData(accelCount);

      MPU9250_Acc[AXIS_X] = (float)accelCount[0] * aRes - accelBias[0];
      MPU9250_Acc[AXIS_Y] = (float)accelCount[1] * aRes - accelBias[1];
      MPU9250_Acc[AXIS_Z] = (float)accelCount[2] * aRes - accelBias[2];
      
      for (uint8_t i = 0; i < AXIS_NUM; i++)
      {
        MPU9250_Acc[i] *= accelScale[i];
      }
     
      readGyroData(gyroCount);

      MPU9250_Gyro[AXIS_X] = (float)gyroCount[0] * gRes - gyroBias[0];
      MPU9250_Gyro[AXIS_Y] = (float)gyroCount[1] * gRes - gyroBias[1];
      MPU9250_Gyro[AXIS_Z] = (float)gyroCount[2] * gRes - gyroBias[2];
      
      for (uint8_t i = 0; i < AXIS_NUM; i++)
      {
        MPU9250_Gyro[i] *= gyroScale[i];
      }
    }
      
    if(readByte(MPU9250_MAGN_I2C_ADDRESS, ST1) & 0x01) 
    {
      readMagData(magCount);
      
      mx = (float)magCount[0] * mRes * magCalibration[0] - magBias[0];
      my = (float)magCount[1] * mRes * magCalibration[1] - magBias[1];
      mz = (float)magCount[2] * mRes * magCalibration[2] - magBias[2];
      
      mx *= magScale[0];
      my *= magScale[1];
      mz *= magScale[2];
    }
    
    Now = uSec;
    deltat = (float)((Now - lastUpdate) / 1000000.0f) ;
    lastUpdate = Now;
        
    sum += deltat;
    sumCount++;
 
    delt_t = HAL_GetTick() - count;
    if (delt_t > 500) 
    {
      tempCount = readTempData();
      temperature = ((float) tempCount) / 333.87f + 21.0f;
      
      MPU9250_CompassAngles[PITCH] *= (-1);
      
      float accVectorLen = sqrt((MPU9250_Acc[AXIS_X] * MPU9250_Acc[AXIS_X]) + 
                                (MPU9250_Acc[AXIS_Y] * MPU9250_Acc[AXIS_Y]) + 
                                (MPU9250_Acc[AXIS_Z] * MPU9250_Acc[AXIS_Z]));
      accPitch = asin((float)MPU9250_Acc[AXIS_X] / accVectorLen) * 180.0f / PI;
      accRoll = asin((float)MPU9250_Acc[AXIS_Y] / accVectorLen) * 180.0f / PI;
      
      MPU9250_CompassAngles[PITCH] = accPitch;
      MPU9250_CompassAngles[ROLL] = accRoll;
      
      float magPitch = -accRoll * PI / 180.0f;
      float magRoll = accPitch * PI / 180.0f;
            
      float xHor = mx * cos(magPitch) +
                   my * sin(magRoll) * sin(magPitch) -
                   mz * cos(magRoll) * sin(magPitch);
      
      float yHor = my * cos(magRoll) + mz * sin(magRoll);
      
      xHorFilt = coef * xHorFilt + (1 - coef) * xHor;
      yHorFilt = coef * yHorFilt + (1 - coef) * yHor;

      heading = atan2(yHorFilt, xHorFilt) * 180.0 / PI;
      
      heading *= (-1);
      heading += 6;
      
      if (heading < 0.0) heading = heading + 360.0;
      
      MPU9250_CompassAngles[YAW] = heading;
   
      count = HAL_GetTick(); 
      sum = 0;
      sumCount = 0;
    }
  }
  else
  {
    if (calibStep >= calibSamplesNum)
    {
      MPU9250_CalibrationFinish();
    }
    else
    {
      uint8_t dataReady = MPU9250_CalibrationStep();
      
      if (dataReady == 1)
      {
        calibStep++;
      }
    }
  }
}



MPU9250_CalibrationStatus MPU9250_GetCalibrationProgress(int* res)
{
  *res = (calibStep * 100) / calibSamplesNum;  
  return calibStatus;
}



/* MPU9250 external library funtions -----------------------------------------*/
void writeByte(uint8_t address, uint8_t subAddress, uint8_t data)
{  
  MPU9250_I2C_Write(address, subAddress, &data, 1);
}



void writeBytes(uint8_t address, uint8_t subAddress, uint8_t count, uint8_t* data)
{  
  for (uint8_t i = 0; i < count; i++)
  {
    MPU9250_I2C_Write(address, subAddress + i, data, 1);
    data++;
  }
}



char readByte(uint8_t address, uint8_t subAddress)
{
  uint8_t data[1];    
  MPU9250_I2C_Read(address, subAddress, data, 1);
  return (char)data[0]; 
}



void readBytes(uint8_t address, uint8_t subAddress, uint8_t count, uint8_t * dest)
{     
  MPU9250_I2C_Read(address, subAddress, dest, count);
}



void getMres() 
{
  switch (Mscale)
  {
    case MFS_14BITS:
      mRes = 10.0 * 4912.0 / 8192.0;
      break;
    case MFS_16BITS:
      mRes = 10.0 * 4912.0 / 32768.0;
      break;
  }
}



void getGres() 
{
  switch (Gscale)
  {
    case GFS_250DPS:
      gRes = 250.0/32768.0;
      break;
      
    case GFS_500DPS:
      gRes = 500.0/32768.0;
      break;
      
    case GFS_1000DPS:
      gRes = 1000.0/32768.0;
      break;
      
    case GFS_2000DPS:
      gRes = 2000.0/32768.0;
      break;
      
    default:
      break;
  }
}



void getAres() 
{
  switch (Ascale)
  {
    case AFS_2G:
      aRes = 2.0/32768.0;
      break;
      
    case AFS_4G:
      aRes = 4.0/32768.0;
      break;
      
    case AFS_8G:
      aRes = 8.0/32768.0;
      break;
      
    case AFS_16G:
      aRes = 16.0/32768.0;
      break;
      
    default:
      break;
  }
}



void readAccelData(int16_t * destination)
{
  uint8_t rawData[6];
  readBytes(MPU9250_I2C_ADDRESS, ACCEL_XOUT_H, 6, &rawData[0]);
  destination[0] = (int16_t)(((int16_t)rawData[0] << 8) | rawData[1]);
  destination[1] = (int16_t)(((int16_t)rawData[2] << 8) | rawData[3]);  
  destination[2] = (int16_t)(((int16_t)rawData[4] << 8) | rawData[5]); 
}



void readGyroData(int16_t * destination)
{
  uint8_t rawData[6];
  readBytes(MPU9250_I2C_ADDRESS, GYRO_XOUT_H, 6, &rawData[0]);
  destination[0] = (int16_t)(((int16_t)rawData[0] << 8) | rawData[1]);
  destination[1] = (int16_t)(((int16_t)rawData[2] << 8) | rawData[3]);  
  destination[2] = (int16_t)(((int16_t)rawData[4] << 8) | rawData[5]); 
}



void readMagData(int16_t * destination)
{
  uint8_t rawData[7];
  
  readBytes(MPU9250_MAGN_I2C_ADDRESS, HXL, 7, &rawData[0]);
  uint8_t c = rawData[6];
  
  if(!(c & 0x08))
  {
    destination[0] = (int16_t)(((int16_t)rawData[1] << 8) | rawData[0]);
    destination[1] = (int16_t)(((int16_t)rawData[3] << 8) | rawData[2]);
    destination[2] = (int16_t)(((int16_t)rawData[5] << 8) | rawData[4]);
  }
}



int16_t readTempData()
{
  uint8_t rawData[2];
  readBytes(MPU9250_I2C_ADDRESS, TEMP_OUT_H, 2, &rawData[0]);
  return (int16_t)(((int16_t)rawData[0]) << 8 | rawData[1]);
}



void resetMPU9250() 
{
  writeByte(MPU9250_I2C_ADDRESS, PWR_MGMT_1, 0x80);
  HAL_Delay(100);
}



void initAK8963(float * destination)
{
  uint8_t rawData[3];
  
  writeByte(MPU9250_MAGN_I2C_ADDRESS, CNTL1, 0x00); 
  HAL_Delay(10);
  
  writeByte(MPU9250_MAGN_I2C_ADDRESS, CNTL1, 0x0F);
  HAL_Delay(10);
  
  readBytes(MPU9250_MAGN_I2C_ADDRESS, ASAX, 3, &rawData[0]);
  destination[0] =  (float)(rawData[0] - 128) / 256.0f + 1.0f;
  destination[1] =  (float)(rawData[1] - 128) / 256.0f + 1.0f;  
  destination[2] =  (float)(rawData[2] - 128) / 256.0f + 1.0f; 
  
  writeByte(MPU9250_MAGN_I2C_ADDRESS, CNTL1, 0x00);
  HAL_Delay(10);
  
  writeByte(MPU9250_MAGN_I2C_ADDRESS, CNTL1, Mscale << 4 | Mmode);
  HAL_Delay(10);
}



void initMPU9250()
{  
  writeByte(MPU9250_I2C_ADDRESS, PWR_MGMT_1, 0x00);
  HAL_Delay(100);

  writeByte(MPU9250_I2C_ADDRESS, PWR_MGMT_1, 0x01);

  writeByte(MPU9250_I2C_ADDRESS, CONFIG_REG, 0x03);  
 
  writeByte(MPU9250_I2C_ADDRESS, SMPLRT_DIV_REG, 0x04);
 
  uint8_t c = readByte(MPU9250_I2C_ADDRESS, GYRO_CONFIG_REG);
  c = c & ~0x02;
  c = c & ~0x18;
  c = c | Gscale << 3;

  writeByte(MPU9250_I2C_ADDRESS, GYRO_CONFIG_REG, c );
  
  c = readByte(MPU9250_I2C_ADDRESS, ACCEL_CONFIG_REG);

  c = c & ~0x18;
  c = c | Ascale << 3;
  writeByte(MPU9250_I2C_ADDRESS, ACCEL_CONFIG_REG, c);

  c = readByte(MPU9250_I2C_ADDRESS, ACCEL_CONFIG_2_REG);
  c = c & ~0x0F;
  c = c | 0x03;
  writeByte(MPU9250_I2C_ADDRESS, ACCEL_CONFIG_2_REG, c);

  writeByte(MPU9250_I2C_ADDRESS, INT_PIN_CFG, 0x22);    
  writeByte(MPU9250_I2C_ADDRESS, INT_ENABLE, 0x01);
}



void MadgwickQuaternionUpdate(float ax, float ay, float az, float gx, float gy, float gz, float mx, float my, float mz)
{
  float q1 = q[0], q2 = q[1], q3 = q[2], q4 = q[3];
  float norm;
  float hx, hy, _2bx, _2bz;
  float s1, s2, s3, s4;
  float qDot1, qDot2, qDot3, qDot4;

  float _2q1mx;
  float _2q1my;
  float _2q1mz;
  float _2q2mx;
  float _4bx;
  float _4bz;
  float _2q1 = 2.0f * q1;
  float _2q2 = 2.0f * q2;
  float _2q3 = 2.0f * q3;
  float _2q4 = 2.0f * q4;
  float _2q1q3 = 2.0f * q1 * q3;
  float _2q3q4 = 2.0f * q3 * q4;
  float q1q1 = q1 * q1;
  float q1q2 = q1 * q2;
  float q1q3 = q1 * q3;
  float q1q4 = q1 * q4;
  float q2q2 = q2 * q2;
  float q2q3 = q2 * q3;
  float q2q4 = q2 * q4;
  float q3q3 = q3 * q3;
  float q3q4 = q3 * q4;
  float q4q4 = q4 * q4;

  norm = sqrt(ax * ax + ay * ay + az * az);
  if (norm == 0.0f) return;
  norm = 1.0f/norm;
  ax *= norm;
  ay *= norm;
  az *= norm;

  norm = sqrt(mx * mx + my * my + mz * mz);
  
  if (norm == 0.0f) 
  {
    return;
  }
  
  norm = 1.0f / norm;
  mx *= norm;
  my *= norm;
  mz *= norm;

  _2q1mx = 2.0f * q1 * mx;
  _2q1my = 2.0f * q1 * my;
  _2q1mz = 2.0f * q1 * mz;
  _2q2mx = 2.0f * q2 * mx;
  hx = mx * q1q1 - _2q1my * q4 + _2q1mz * q3 + mx * q2q2 + _2q2 * my * q3 + _2q2 * mz * q4 - mx * q3q3 - mx * q4q4;
  hy = _2q1mx * q4 + my * q1q1 - _2q1mz * q2 + _2q2mx * q3 - my * q2q2 + my * q3q3 + _2q3 * mz * q4 - my * q4q4;
  _2bx = sqrt(hx * hx + hy * hy);
  _2bz = -_2q1mx * q3 + _2q1my * q2 + mz * q1q1 + _2q2mx * q4 - mz * q2q2 + _2q3 * my * q4 - mz * q3q3 + mz * q4q4;
  _4bx = 2.0f * _2bx;
  _4bz = 2.0f * _2bz;

  s1 = -_2q3 * (2.0f * q2q4 - _2q1q3 - ax) + _2q2 * (2.0f * q1q2 + _2q3q4 - ay) - _2bz * q3 * (_2bx * (0.5f - q3q3 - q4q4) + _2bz * (q2q4 - q1q3) - mx) + (-_2bx * q4 + _2bz * q2) * (_2bx * (q2q3 - q1q4) + _2bz * (q1q2 + q3q4) - my) + _2bx * q3 * (_2bx * (q1q3 + q2q4) + _2bz * (0.5f - q2q2 - q3q3) - mz);
  s2 = _2q4 * (2.0f * q2q4 - _2q1q3 - ax) + _2q1 * (2.0f * q1q2 + _2q3q4 - ay) - 4.0f * q2 * (1.0f - 2.0f * q2q2 - 2.0f * q3q3 - az) + _2bz * q4 * (_2bx * (0.5f - q3q3 - q4q4) + _2bz * (q2q4 - q1q3) - mx) + (_2bx * q3 + _2bz * q1) * (_2bx * (q2q3 - q1q4) + _2bz * (q1q2 + q3q4) - my) + (_2bx * q4 - _4bz * q2) * (_2bx * (q1q3 + q2q4) + _2bz * (0.5f - q2q2 - q3q3) - mz);
  s3 = -_2q1 * (2.0f * q2q4 - _2q1q3 - ax) + _2q4 * (2.0f * q1q2 + _2q3q4 - ay) - 4.0f * q3 * (1.0f - 2.0f * q2q2 - 2.0f * q3q3 - az) + (-_4bx * q3 - _2bz * q1) * (_2bx * (0.5f - q3q3 - q4q4) + _2bz * (q2q4 - q1q3) - mx) + (_2bx * q2 + _2bz * q4) * (_2bx * (q2q3 - q1q4) + _2bz * (q1q2 + q3q4) - my) + (_2bx * q1 - _4bz * q3) * (_2bx * (q1q3 + q2q4) + _2bz * (0.5f - q2q2 - q3q3) - mz);
  s4 = _2q2 * (2.0f * q2q4 - _2q1q3 - ax) + _2q3 * (2.0f * q1q2 + _2q3q4 - ay) + (-_4bx * q4 + _2bz * q2) * (_2bx * (0.5f - q3q3 - q4q4) + _2bz * (q2q4 - q1q3) - mx) + (-_2bx * q1 + _2bz * q3) * (_2bx * (q2q3 - q1q4) + _2bz * (q1q2 + q3q4) - my) + _2bx * q2 * (_2bx * (q1q3 + q2q4) + _2bz * (0.5f - q2q2 - q3q3) - mz);
  norm = sqrt(s1 * s1 + s2 * s2 + s3 * s3 + s4 * s4);
  norm = 1.0f/norm;
  s1 *= norm;
  s2 *= norm;
  s3 *= norm;
  s4 *= norm;

  qDot1 = 0.5f * (-q2 * gx - q3 * gy - q4 * gz) - beta * s1;
  qDot2 = 0.5f * (q1 * gx + q3 * gz - q4 * gy) - beta * s2;
  qDot3 = 0.5f * (q1 * gy - q2 * gz + q4 * gx) - beta * s3;
  qDot4 = 0.5f * (q1 * gz + q2 * gy - q3 * gx) - beta * s4;

  q1 += qDot1 * deltat;
  q2 += qDot2 * deltat;
  q3 += qDot3 * deltat;
  q4 += qDot4 * deltat;
  norm = sqrt(q1 * q1 + q2 * q2 + q3 * q3 + q4 * q4);
  norm = 1.0f/norm;
  q[0] = q1 * norm;
  q[1] = q2 * norm;
  q[2] = q3 * norm;
  q[3] = q4 * norm;
}



void MahonyQuaternionUpdate(float ax, float ay, float az, float gx, float gy, float gz, float mx, float my, float mz)
{
  float q1 = q[0], q2 = q[1], q3 = q[2], q4 = q[3];
  float norm;
  float hx, hy, bx, bz;
  float vx, vy, vz, wx, wy, wz;
  float ex, ey, ez;
  float pa, pb, pc;

  float q1q1 = q1 * q1;
  float q1q2 = q1 * q2;
  float q1q3 = q1 * q3;
  float q1q4 = q1 * q4;
  float q2q2 = q2 * q2;
  float q2q3 = q2 * q3;
  float q2q4 = q2 * q4;
  float q3q3 = q3 * q3;
  float q3q4 = q3 * q4;
  float q4q4 = q4 * q4;   

  norm = sqrt(ax * ax + ay * ay + az * az);
  
  if (norm == 0.0f)
  {
    return;
  }
  
  norm = 1.0f / norm;
  ax *= norm;
  ay *= norm;
  az *= norm;

  norm = sqrt(mx * mx + my * my + mz * mz);
  
  if (norm == 0.0f)
  {
    return;
  }
  
  norm = 1.0f / norm;
  mx *= norm;
  my *= norm;
  mz *= norm;

  hx = 2.0f * mx * (0.5f - q3q3 - q4q4) + 2.0f * my * (q2q3 - q1q4) + 2.0f * mz * (q2q4 + q1q3);
  hy = 2.0f * mx * (q2q3 + q1q4) + 2.0f * my * (0.5f - q2q2 - q4q4) + 2.0f * mz * (q3q4 - q1q2);
  bx = sqrt((hx * hx) + (hy * hy));
  bz = 2.0f * mx * (q2q4 - q1q3) + 2.0f * my * (q3q4 + q1q2) + 2.0f * mz * (0.5f - q2q2 - q3q3);

  vx = 2.0f * (q2q4 - q1q3);
  vy = 2.0f * (q1q2 + q3q4);
  vz = q1q1 - q2q2 - q3q3 + q4q4;
  wx = 2.0f * bx * (0.5f - q3q3 - q4q4) + 2.0f * bz * (q2q4 - q1q3);
  wy = 2.0f * bx * (q2q3 - q1q4) + 2.0f * bz * (q1q2 + q3q4);
  wz = 2.0f * bx * (q1q3 + q2q4) + 2.0f * bz * (0.5f - q2q2 - q3q3);  

  ex = (ay * vz - az * vy) + (my * wz - mz * wy);
  ey = (az * vx - ax * vz) + (mz * wx - mx * wz);
  ez = (ax * vy - ay * vx) + (mx * wy - my * wx);
  
  if (Ki > 0.0f)
  {
    eInt[0] += ex;
    eInt[1] += ey;
    eInt[2] += ez;
  }
  else
  {
    eInt[0] = 0.0f;
    eInt[1] = 0.0f;
    eInt[2] = 0.0f;
  }

  gx = gx + Kp * ex + Ki * eInt[0];
  gy = gy + Kp * ey + Ki * eInt[1];
  gz = gz + Kp * ez + Ki * eInt[2];

  pa = q2;
  pb = q3;
  pc = q4;
  q1 = q1 + (-q2 * gx - q3 * gy - q4 * gz) * (0.5f * deltat);
  q2 = pa + (q1 * gx + pb * gz - pc * gy) * (0.5f * deltat);
  q3 = pb + (q1 * gy - pa * gz + pc * gx) * (0.5f * deltat);
  q4 = pc + (q1 * gz + pa * gy - pb * gx) * (0.5f * deltat);

  norm = sqrt(q1 * q1 + q2 * q2 + q3 * q3 + q4 * q4);
  norm = 1.0f / norm;
  q[0] = q1 * norm;
  q[1] = q2 * norm;
  q[2] = q3 * norm;
  q[3] = q4 * norm;
}