/**
  ******************************************************************************
  * @file           : mpu9250.h
  * @brief          : MPU9250 driver header
  * @author         : MicroTechnics (microtechnics.ru)
  ******************************************************************************
  */

#ifndef MPU9250_H
#define MPU9250_H



/* Includes ------------------------------------------------------------------*/
    
#include "stm32f1xx_hal.h"
#include "math.h"
#include "stdio.h"
#include "flash.h"



/* Declarations and definitions ----------------------------------------------*/

// MPU9250 registers' addresses
#define SELF_TEST_X_GYRO_REG                                            0x00
#define SELF_TEST_Y_GYRO_REG                                            0x01
#define SELF_TEST_Z_GYRO_REG                                            0x02
#define SELF_TEST_X_ACCEL_REG                                           0x0D
#define SELF_TEST_Y_ACCEL_REG                                           0x0E
#define SELF_TEST_Z_ACCEL_REG                                           0x0F
#define XG_OFFSET_H_REG                                                 0x13
#define XG_OFFSET_L_REG                                                 0x14
#define YG_OFFSET_H_REG                                                 0x15
#define YG_OFFSET_L_REG                                                 0x16
#define ZG_OFFSET_H_REG                                                 0x17
#define ZG_OFFSET_L_REG                                                 0x18
#define SMPLRT_DIV_REG                                                  0x19
#define CONFIG_REG                                                      0x1A
#define GYRO_CONFIG_REG                                                 0x1B
#define ACCEL_CONFIG_REG                                                0x1C
#define ACCEL_CONFIG_2_REG                                              0x1D
#define LP_ACCEL_ODR_REG                                                0x1E
#define WOM_THR_REG                                                     0x1F
#define FIFO_EN_REG                                                     0x23
#define I2C_MST_CTRL_REG                                                0x24
#define I2C_SLV0_ADDR_REG                                               0x25
#define I2C_SLV0_REG_REG                                                0x26
#define I2C_SLV0_CTRL_REG                                               0x27
#define I2C_SLV1_ADDR_REG                                               0x28
#define I2C_SLV1_REG_REG                                                0x29
#define I2C_SLV1_CTRL _REG                                              0x2A
#define I2C_SLV2_ADDR_REG                                               0x2B
#define I2C_SLV2_REG_REG                                                0x2C
#define I2C_SLV2_CTRL_REG                                               0x2D
#define I2C_SLV3_ADDR                                                   0x2E
#define I2C_SLV3_REG                                                    0x2F
#define I2C_SLV3_CTRL                                                   0x30
#define I2C_SLV4_ADDR                                                   0x31
#define I2C_SLV4_REG                                                    0x32
#define I2C_SLV4_DO                                                     0x33
#define I2C_SLV4_CTRL                                                   0x34
#define I2C_SLV4_DI                                                     0x35
#define I2C_MST_STATUS                                                  0x36
#define INT_PIN_CFG                                                     0x37
#define INT_ENABLE                                                      0x38
#define INT_STATUS                                                      0x3A
#define ACCEL_XOUT_H                                                    0x3B
#define ACCEL_XOUT_L                                                    0x3C
#define ACCEL_YOUT_H                                                    0x3D
#define ACCEL_YOUT_L                                                    0x3E
#define ACCEL_ZOUT_H                                                    0x3F
#define ACCEL_ZOUT_L                                                    0x40
#define TEMP_OUT_H                                                      0x41
#define TEMP_OUT_L                                                      0x42
#define GYRO_XOUT_H                                                     0x43
#define GYRO_XOUT_L                                                     0x44
#define GYRO_YOUT_H                                                     0x45
#define GYRO_YOUT_L                                                     0x46
#define GYRO_ZOUT_H                                                     0x47
#define GYRO_ZOUT_L                                                     0x48
#define EXT_SENS_DATA_00                                                0x49
#define EXT_SENS_DATA_01                                                0x4A
#define EXT_SENS_DATA_02                                                0x4B
#define EXT_SENS_DATA_03                                                0x4C
#define EXT_SENS_DATA_04                                                0x4D
#define EXT_SENS_DATA_05                                                0x4E
#define EXT_SENS_DATA_06                                                0x4F
#define EXT_SENS_DATA_07                                                0x50
#define EXT_SENS_DATA_08                                                0x51
#define EXT_SENS_DATA_09                                                0x52
#define EXT_SENS_DATA_10                                                0x53
#define EXT_SENS_DATA_11                                                0x54
#define EXT_SENS_DATA_12                                                0x55
#define EXT_SENS_DATA_13                                                0x56
#define EXT_SENS_DATA_14                                                0x57
#define EXT_SENS_DATA_15                                                0x58
#define EXT_SENS_DATA_16                                                0x59
#define EXT_SENS_DATA_17                                                0x5A
#define EXT_SENS_DATA_18                                                0x5B
#define EXT_SENS_DATA_19                                                0x5C
#define EXT_SENS_DATA_20                                                0x5D
#define EXT_SENS_DATA_21                                                0x5E
#define EXT_SENS_DATA_22                                                0x5F
#define EXT_SENS_DATA_23                                                0x60
#define I2C_SLV0_DO                                                     0x63
#define I2C_SLV1_DO                                                     0x64
#define I2C_SLV2_DO                                                     0x65
#define I2C_SLV3_DO                                                     0x66
#define I2C_MST_DELAY_CTRL                                              0x67
#define SIGNAL_PATH_RESET                                               0x68
#define MOT_DETECT_CTRL                                                 0x69
#define USER_CTRL                                                       0x6A
#define PWR_MGMT_1                                                      0x6B
#define PWR_MGMT_2                                                      0x6C
#define FIFO_COUNTH                                                     0x72
#define FIFO_COUNTL                                                     0x73
#define FIFO_R_W                                                        0x74
#define WHO_AM_I                                                        0x75
#define XA_OFFSET_H                                                     0x77
#define XA_OFFSET_L                                                     0x78
#define YA_OFFSET_H                                                     0x7A
#define YA_OFFSET_L                                                     0x7B
#define ZA_OFFSET_H                                                     0x7D
#define ZA_OFFSET_L                                                     0x7E
#define WIA                                                             0x00
#define INFO                                                            0x01
#define ST1                                                             0x02
#define HXL                                                             0x03
#define HXH                                                             0x04
#define HYL                                                             0x05
#define HYH                                                             0x06
#define HZL                                                             0x07
#define HZH                                                             0x08
#define ST2                                                             0x09
#define CNTL1                                                           0x0A
#define CNTL2                                                           0x0B
#define ASTC                                                            0x0C
#define TS1                                                             0x0D
#define TS2                                                             0x0E
#define I2CDIS                                                          0x0F
#define ASAX                                                            0x10
#define ASAY                                                            0x11
#define ASAZ                                                            0x12

#define MPU9250_I2C_ADDRESS                                             0x68
#define MPU9250_I2C_TIMEOUT                                             25
#define MPU9250_I2C_MAX_DATA_LEN                                        16
#define MPU9250_MAGN_I2C_ADDRESS                                        0x0C

#define MPU9250_MAG_CALIB_STEPS_NUM                                     2400
#define MPU9250_ACC_CALIB_STEPS_NUM                                     120
#define MPU9250_GYRO_CALIB_STEPS_NUM                                    120

typedef enum
{
  MPU9250_OK              = 0x00,
  MPU9250_ERROR           = 0x01,
  MPU9250_I2C_OVERFLOW    = 0x02,
} MPU9250_Status;

typedef enum
{
  MPU9250_CALIBRATION_NOT_ACTIVE = 0x00,
  MPU9250_MAG_CALIBRATION        = 0x01,
  MPU9250_ACC_CALIBRATION        = 0x02,
  MPU9250_GYRO_CALIBRATION       = 0x03,
} MPU9250_CalibrationStatus;    

enum CompassAngle {YAW, PITCH, ROLL, ANGLES_NUM};
enum Axis {AXIS_X, AXIS_Y, AXIS_Z, AXIS_NUM};

extern float MPU9250_CompassAngles[ANGLES_NUM];
extern float MPU9250_Acc[AXIS_NUM];
extern float MPU9250_Gyro[AXIS_NUM];



/* Functions -----------------------------------------------------------------*/

extern void MPU9250_Init();
extern void MPU9250_Process();
extern void MPU9250_CalibrationStart(MPU9250_Sensor sensor);
extern MPU9250_CalibrationStatus MPU9250_GetCalibrationProgress(int* res);



#endif // MPU9250_H 